#include "onedataset.h"
#include "generate.h"
#include "suh.h"
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "zerofinder.h"
#include "fitfreq.h"
#include "accumulate.h"

const int est=1;

double huge_mu = 10000000000.0;
double tiny_mu = .0000000001;
double tiny;


extern int numruns,numallele;
extern int *allelesize;
extern double *expge;
extern int *replicates;
extern int **tableofalleles;
extern int progenitors[2];

extern _seen seen;
extern _seen unseen;
extern _freq freq;
extern _mu mu;
extern _mutant mutant;
extern _calibrate calibrate;


void fitone(int bootstrap, int writesimulation,int simunit)
{
	int x;
	//modifydata();
	if(bootstrap)
	{
		fitallmu(seen.boot,unseen.boot,mu.init_est.boot,mu.est.boot,mu.var.boot);
		for(x=0;x<numallele;x++)
			mu.sd.boot[x]= sqrt(mu.var.boot[x]);
		est_var_freq(mu.est.boot,mu.var.boot,freq.est.untrans.boot,freq.var.untrans.boot,freq.sd.untrans.boot);
		for(x=0;x<numallele;x++)
			freq.est.trans.boot[x]=asin(freq.est.untrans.boot[x]);  // assuming asin and arcsin are the some thing
		for(x=0;x<numallele;x++)
			freq.var.trans.boot[x]=asy_var_arcsin(freq.est.untrans.boot[x],freq.var.untrans.boot[x]);
		for(x=0;x<numallele;x++)
			freq.sd.trans.boot[x]=sqrt(freq.var.trans.boot[x]);
		est_var_calibrate(mu.est.boot,mu.var.boot,&calibrate.est.boot,&calibrate.var.boot,&calibrate.sd.boot);
		est_var_mut(mu.est.boot,mu.var.boot,&mutant.est.untrans.boot,&mutant.var.untrans.boot,&mutant.sd.untrans.boot);
		mutant.est.trans.boot=asin(mutant.est.untrans.boot);
		mutant.var.trans.boot=asy_var_arcsin(mutant.est.untrans.boot,mutant.var.untrans.boot);
		mutant.sd.trans.boot=sqrt(mutant.var.trans.boot);
		// bootstrap specific
		for(x=0;x<numallele;x++)
			accumulate(mu.est.boot[x],&mu.esum.boot[x],&mu.ess.boot[x]);
		for(x=0;x<numallele;x++)
			accumulate(freq.est.untrans.boot[x],&freq.esum.untrans.boot[x],&freq.ess.trans.boot[x]);
		for(x=0;x<numallele;x++)
			accumulate(freq.est.trans.boot[x],&freq.esum.trans.boot[x],&freq.ess.trans.boot[x]);
		accumulate(calibrate.est.boot,&calibrate.esum.boot,&calibrate.ess.boot);
		accumulate(mutant.est.untrans.boot,&mutant.esum.untrans.boot,&mutant.ess.untrans.boot);
		accumulate(mutant.est.trans.boot,&mutant.esum.trans.boot,&mutant.ess.trans.boot);
		if(writesimulation)
		{
			printf("Cal: %10.4lf\n",calibrate.est.boot);
			printf("Mu: ");
			for(x=0;x<numallele;x++)
				printf("%8.4lf  ",mu.est.boot[x]);
			printf("Freq: ");
				printf("%8.4lf  ",freq.est.untrans.boot[x]);
		}
	}
	else
	{
		fitallmu(seen.data,unseen.data,mu.init_est.data,mu.est.data,mu.var.data);
		//printf("inital estimates\n");
		//for(x=0;x<numallele;x++)
		//	printf("%.8lf ",mu.init_est.data[x]);
		//printf("\n");
		//printf("final estimates\n");
		//for(x=0;x<numallele;x++)
		//	printf("%.8lf ",mu.est.data[x]);
		//printf("\n");
		for(x=0;x<numallele;x++)
			mu.sd.data[x]= sqrt(mu.var.data[x]);
		//printf("estimated variance frequency\n");
		est_var_freq(mu.est.data,mu.var.data,freq.est.untrans.data,freq.var.untrans.data,freq.sd.untrans.data);	
		//printf("arcsin frequencny\n");
		for(x=0;x<numallele;x++)
			freq.est.trans.data[x]=asin(freq.est.untrans.data[x]);   // assuming asin and arcsin are the some thing
		//printf("asymptotic variance arcsin\n");
		for(x=0;x<numallele;x++)
			freq.var.trans.data[x]=asy_var_arcsin(freq.est.untrans.data[x],freq.var.untrans.data[x]);	
		for(x=0;x<numallele;x++)
			freq.sd.trans.data[x]=sqrt(freq.var.trans.data[x]);
		//printf("calibrate\n");
		est_var_calibrate(mu.est.data,mu.var.data,&calibrate.est.data,&calibrate.var.data,&calibrate.sd.data);
		//printf("estimated variance mutation\n");
		est_var_mut(mu.est.data,mu.var.data,&mutant.est.untrans.data,&mutant.var.untrans.data,&mutant.sd.untrans.data);
		mutant.est.trans.data=asin(mutant.est.untrans.data);
		mutant.var.trans.data=asy_var_arcsin(mutant.est.untrans.data,mutant.var.untrans.data);
		mutant.sd.trans.data=sqrt(mutant.var.trans.data);
		//printf("done with dataset\n");
	}

}

void bootstrap()
{
	int x;
	printf("generating  set\n");
	set_generate();
	printf("doing bootstrap: \n");
	for(x=0;x<num_repeats;x++)
	{
		generate();
		fitone(true,0,0);
	}
	printf("\n");
	boot_answers();
}

void modifydata()
{
	int x,y;
	int ok=1;
	// see if unseen is 0 anywhere
	for(x=0;x<numruns;x++)
	{
		//if(unseen[x]==0)
		//	ok=0;
	}
	if(ok)
		return;
	// unseen zero is somewhere, make modifications
	for(y=0;y<numruns;y++)
	{
		//if(unseen[y]>0)
		//	continue;
		//else
		{
			
		
		}
	}


}

void fitallmu(double seen[50][50],double unseen[50][50],double *mu_init_est, double *mu_final_est, double *mu_est_var)
{
	int x;
	/*for(x=0;x<numallele;x++)
	{
		for(y=0;y<numruns;y++)
			printf("%4.0lf ",seen[x][y]);
		printf("\n");
	
	}*/
	for(x=0;x<numallele;x++)
	{
		fitonemu(seen[x],unseen[x],&mu_init_est[x],&mu_final_est[x],&mu_est_var[x]);
	}

}

void fitonemu(double *seen,double *unseen, double *mu_init_est, double *mu_final,double *mu_est_var)
{
	*mu_init_est=initialestimate(expge,seen,unseen);
	//printf("initial estimate: %lf\n",*mu_init_est);
	*mu_final=refineestimate(expge,seen,unseen,*mu_init_est);
	//printf("refined estimate: %lf\n",*mu_final);
	ll_derivs(expge,*mu_final,seen,unseen,4,NULL,NULL,mu_est_var);
	if(*mu_est_var<0)
	{
		*mu_est_var = -1.0/(*mu_est_var);
	}
	else
		*mu_est_var=0;//*/
}
double initialestimate(double *dna, double *seen,double *unseen)
{
	int x;
	double small_punseen = .0001;
	double punseen;
	double *temp;
	temp=(double*)malloc(sizeof(double)*numruns);
	/*for(x=0;x<numruns;x++)
	{
		printf("dna[%d]=%lf ",x,dna[x]);
	}
	printf("\n");*/
	for(x=0;x<numruns;x++)
	{
		if(seen[x]+unseen[x]==0)
		{
			printf("seen[x]+unseen[x]==0");
			exit(0);
		}
		temp[x]=unseen[x]/(seen[x]+unseen[x]);
		//printf("%lf/(%lf+%lf)[%lf] = temp[x]=%lf\n",unseen[x],seen[x],unseen[x],unseen[x]+seen[x],temp[x]);
	}
	punseen=small_punseen;
	//printf("punseen=%lf\n",punseen);
	for(x=0;x<numruns;x++)
	{
		if(temp[x]>punseen)
			punseen=temp[x];
		if(temp[x]<small_punseen)		// kill a 0
			temp[x]=small_punseen;
	}
	//printf("punseen=%lf\n",punseen);
	if(punseen==small_punseen)
	{
		//printf("punseen = small_punseen\n");
		for(x=0;x<numruns;x++)
		{
			temp[x]=-log(punseen);
			//printf("temp[%d]=%lf ",x,temp[x]);
		}
		//printf("\n");
	}
	else
	{
		//printf("punseen != small_punseen\n");
		for(x=0;x<numruns;x++)
		{
			temp[x]=-log(temp[x]);
			//printf("temp[%d]=%lf ",x,temp[x]);
		}
		//printf("\n");
	}
	for(x=0;x<numruns;x++)
	{
		//printf("%d) %lf",x,temp[x]);
		temp[x]/=dna[x];
		//printf("/=%lf = %lf  ",dna[x],temp[x]);
	}
	//printf("\n");
	punseen=0;
	for(x=0;x<numruns;x++)
	{
		//printf("temp[%d]=%lf ",x,temp[x]);
		punseen+=temp[x];
	}
	punseen/=numruns;
	//printf("\npunseen=%lf\n",punseen);
	free(temp);
	return punseen;
}

double refineestimate(double *dna, double *seen, double *unseen, double mu_init_est)
{
	int x;
	double sum=0;
	double est;
	int status;
	double deriv = 0;
	double *localseen,*localunseen;
	localseen=(double*)malloc(sizeof(double)*numruns);
	localunseen=(double*)malloc(sizeof(double)*numruns);
	for(x=0;x<numruns;x++)
	{
		localunseen[x]=unseen[x];
		localseen[x]=seen[x];
	}
	tiny= pow(10,-100);
	for(x=0;x<numruns;x++)
		sum+=seen[x];
	if(sum<tiny_mu) // REALLY small seen, then fudge data to assure something is seen
	{
		localseen[0]+=.01;
		localunseen[0]-=.01;
	}
	setZeroFinder(0,huge_mu);
	est = mu_init_est;
	status=0;
	//printf("*****starting a refinement estimation*****\n");
	while(1)
	{
			rc_step_zf(&status,&est,deriv);
			//printf("est=%lf\n",est);
			if(status!=1)
			{
				break;
			}
			ll_derivs(dna,est,localseen,localunseen,2,NULL,&deriv,NULL);
			
	}
	
	free(localunseen);
	free(localseen);
	
	return est;
}

double expm1(double x)
{
	if (fabs(x) < 1e-5)
		return x + 0.5*x*x;
	else
		return exp(x) - 1.0;
}


// optional parameters, 
// opt = 1, do ll
// opt = 2, do dll_dmu
// opt = 4, do d2ll_dmu2
// and of course all combination of these bits
void ll_derivs(double *dna, double mu, double *seen, double *unseen,int opt,double *ll,double *dll_mu, double *d2ll_dmu2)
{
	int x;
	double max;
	double *mean=(double*)malloc(sizeof(double)*numruns);
	double *pseen=(double*)malloc(sizeof(double)*numruns);
	double *punseen=(double*)malloc(sizeof(double)*numruns);
	double *pdunseen=(double*)malloc(sizeof(double)*numruns);
	for(x=0;x<numruns;x++)
	{
		mean[x]=mu*dna[x];
		//printf("mean[%d]=%lf ",x,mean[x]);
	}
	//printf("\n");
	for(x=0;x<numruns;x++)
	{
		pseen[x]=-expm1(-mean[x]);
		//printf("pseen[%d]=%.12lf ",x,pseen[x]);
	}
	//printf("\n");
	for(x=0;x<numruns;x++)
	{
		if(pseen[x]<tiny)
		{
			pseen[x]=tiny;
		}
	}
	for(x=0;x<numruns;x++)
	{
		punseen[x]=exp(-mean[x]);
		//printf("punseen[%d]=%lf ",x,punseen[x]);
	}
	//printf("\n");
	for(x=0;x<numruns;x++)
		pdunseen[x]=punseen[x]/pseen[x];
	if(opt&1)
	{
		*ll=0;
		for(x=0;x<numruns;x++)
			*ll+=(seen[x]*log(pseen[x]))-(unseen[x]*mean[x]);
	}
	if(opt&2)
	{
		*dll_mu=0;
		for(x=0;x<numruns;x++)
		{
			*dll_mu+=dna[x]*(seen[x]*pdunseen[x]-unseen[x]);
			//printf("expge[%d](%.4lf) * (seen[%d](%.4lf) * pdunseen[%d](%.4lf) - unseen[%d](%.4lf) = %.4lf\n",x,expge[x],x,seen[x],x,pdunseen[x],x,unseen[x],*dll_mu);
		}
	}
	if(opt&4)
	{
		*d2ll_dmu2=0;
		for(x=0;x<numruns;x++)
		{
			*d2ll_dmu2+=seen[x]*pow(dna[x],2)*pdunseen[x]/pseen[x];
			//printf("seen[%d](%lf) * %lf * pdunseen[%d](%lf) / pseen[%d](%lf) = %lf\n",x,seen[x],pow(dna[x],2),x,pdunseen[x],x,pseen[x],*d2ll_dmu2);
		}
		*d2ll_dmu2=-*d2ll_dmu2;
	
	}
}

double oexpm1(double x)
{
	double p1=.914041914819518*pow(10,-9);
	double p2=.238082361044469*pow(10,-1);
	double q1=-.499999999085958;
	double q2=.107141568980644;
	double q3=-.119041179760821*pow(10,-1);
	double q4=.595130811860248*pow(10,-3);
	double w;
	if(fabs(x)<=.15)
	{
		return  x*(((p2*x+p1)*x+1.0)/((((q4*x+q3)*x+q2)*x+q1)*x+1.0));
	}
	w=exp(x);
	if(x<=0)
		return (w-.5)-.5;
	
	return w*(.5 + (.5 - 1.0/w));

}

double asy_var_arcsin(double p, double var_p)   // asymptotic variance of the arcsin transform of p
{
	//double tiny = pow(10,-10);
	double localp;
	localp=p>tiny?p:tiny;		// worthless statement
	localp=localp<1-tiny?localp:1-tiny;
	return var_p/(localp*(1-localp));
	

}
double inverse_arcsin(double x)
{
	return pow(sin(x/2.0),2);

}

// leave room here for his wacktacular arcsin function